@@ -1,5 +1,5 @@
 /* stat.c -- display file or filesystem status
-   Copyright (C) 2001, 2002 Free Software Foundation.
+   Copyright (C) 2001, 2002, 2003 Free Software Foundation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -43,18 +43,17 @@
 # endif
 #endif
 
-#include <string.h>
-#include <ctype.h>
-
 #include "system.h"
 
 #include "closeout.h"
 #include "error.h"
 #include "filemode.h"
+#include "file-type.h"
 #include "fs.h"
 #include "getopt.h"
 #include "quote.h"
 #include "quotearg.h"
+#include "strftime.h"
 #include "xreadlink.h"
 
 #define NAMEMAX_FORMAT PRIuMAX
@@ -80,9 +79,6 @@
 # define NAMEMAX_FORMAT "s"
 #endif
 
-size_t nstrftime PARAMS ((char *, size_t, char const *,
-			  struct tm const *, int, int));
-
 #define PROGRAM_NAME "stat"
 
 #define AUTHORS "Michael Meskes"
@@ -98,40 +94,10 @@ static struct option const long_options[] = {
   {NULL, 0, NULL, 0}
 };
 
-char *program_name;
+/* Nonzero means we should exit with EXIT_FAILURE upon completion.  */
+static int G_fail;
 
-static void
-print_human_type (mode_t mode)
-{
-  char const *type;
-  switch (mode & S_IFMT)
-    {
-    case S_IFDIR:
-      type = "Directory";
-      break;
-    case S_IFCHR:
-      type = "Character Device";
-      break;
-    case S_IFBLK:
-      type = "Block Device";
-      break;
-    case S_IFREG:
-      type = "Regular File";
-      break;
-    case S_IFLNK:
-      type = "Symbolic Link";
-      break;
-    case S_IFSOCK:
-      type = "Socket";
-      break;
-    case S_IFIFO:
-      type = "Fifo File";
-      break;
-    default:
-      type = "Unknown";
-    }
-  fputs (type, stdout);
-}
+char *program_name;
 
 /* Return the type of the specified file system.
    Some systems have statfvs.f_basetype[FSTYPSZ]. (AIX, HP-UX, and Solaris)
@@ -339,29 +305,27 @@ human_fstype (STRUCT_STATVFS const *statfsbuf)
 #endif
 }
 
-static void
-print_human_access (struct stat const *statbuf)
+static char *
+human_access (struct stat const *statbuf)
 {
-  char modebuf[11];
+  static char modebuf[11];
   mode_string (statbuf->st_mode, modebuf);
   modebuf[10] = 0;
-  fputs (modebuf, stdout);
+  return modebuf;
 }
 
-static void
-print_human_time (time_t const *t)
+static char *
+human_time (time_t const *t)
 {
-  char str[80];
-
-#if 0  /* %c is too locale-dependent.  */
-  if (strftime (str, 40, "%c", localtime (t)) > 0)
-#else
-  if (nstrftime (str, sizeof str, "%Y-%m-%d %H:%M:%S.%N %z",
-		 localtime (t), 0, 0) > 0)
-#endif
-    fputs (str, stdout);
-  else
-    printf ("Cannot calculate human readable time, sorry");
+  static char str[80];
+  struct tm *tm = localtime (t);
+  if (tm == NULL)
+    {
+      G_fail = 1;
+      return (char *) _("*** invalid date/time ***");
+    }
+  nstrftime (str, sizeof str, "%Y-%m-%d %H:%M:%S.%N %z", tm, 0, 0);
+  return str;
 }
 
 /* print statfs info */
@@ -453,7 +417,7 @@ print_stat (char *pformat, char m, char const *filename, void const *data)
       break;
     case 'N':
       strcat (pformat, "s");
-      if ((statbuf->st_mode & S_IFMT) == S_IFLNK)
+      if (S_ISLNK (statbuf->st_mode))
 	{
 	  char *linkname = xreadlink (filename);
 	  if (linkname == NULL)
@@ -489,14 +453,16 @@ print_stat (char *pformat, char m, char const *filename, void const *data)
       printf (pformat, statbuf->st_mode & 07777);
       break;
     case 'A':
-      print_human_access (statbuf);
+      strcat (pformat, "s");
+      printf (pformat, human_access (statbuf));
       break;
     case 'f':
       strcat (pformat, "x");
       printf (pformat, statbuf->st_mode);
       break;
     case 'F':
-      print_human_type (statbuf->st_mode);
+      strcat (pformat, "s");
+      printf (pformat, file_type (statbuf));
       break;
     case 'h':
       strcat (pformat, "d");
@@ -536,28 +502,31 @@ print_stat (char *pformat, char m, char const *filename, void const *data)
       break;
     case 'b':
       strcat (pformat, "u");
-      printf (pformat, (unsigned int) statbuf->st_blocks);
+      printf (pformat, (unsigned int) ST_NBLOCKS (*statbuf));
       break;
     case 'o':
       strcat (pformat, "d");
       printf (pformat, (int) statbuf->st_blksize);
       break;
     case 'x':
-      print_human_time (&(statbuf->st_atime));
+      strcat (pformat, "s");
+      printf (pformat, human_time (&(statbuf->st_atime)));
       break;
     case 'X':
       strcat (pformat, "d");
       printf (pformat, (int) statbuf->st_atime);
       break;
     case 'y':
-      print_human_time (&(statbuf->st_mtime));
+      strcat (pformat, "s");
+      printf (pformat, human_time (&(statbuf->st_mtime)));
       break;
     case 'Y':
       strcat (pformat, "d");
       printf (pformat, (int) statbuf->st_mtime);
       break;
     case 'z':
-      print_human_time (&(statbuf->st_ctime));
+      strcat (pformat, "s");
+      printf (pformat, human_time (&(statbuf->st_ctime)));
       break;
     case 'Z':
       strcat (pformat, "d");
@@ -722,56 +691,56 @@ Display file or filesystem status.\n\
       fputs (_("\n\
 The valid format sequences for files (without --filesystem):\n\
 \n\
-  %A - Access rights in human readable form\n\
-  %a - Access rights in octal\n\
-  %b - Number of blocks allocated\n\
+  %A   Access rights in human readable form\n\
+  %a   Access rights in octal\n\
+  %b   Number of blocks allocated\n\
 "), stdout);
       fputs (_("\
-  %D - Device number in hex\n\
-  %d - Device number in decimal\n\
-  %F - File type\n\
-  %f - raw mode in hex\n\
-  %G - Group name of owner\n\
-  %g - Group ID of owner\n\
+  %D   Device number in hex\n\
+  %d   Device number in decimal\n\
+  %F   File type\n\
+  %f   raw mode in hex\n\
+  %G   Group name of owner\n\
+  %g   Group ID of owner\n\
 "), stdout);
       fputs (_("\
-  %h - Number of hard links\n\
-  %i - Inode number\n\
-  %N - Quoted File name with dereference if symbolic link\n\
-  %n - File name\n\
-  %o - IO block size\n\
-  %s - Total size, in bytes\n\
-  %T - Minor device type in hex\n\
-  %t - Major device type in hex\n\
+  %h   Number of hard links\n\
+  %i   Inode number\n\
+  %N   Quoted File name with dereference if symbolic link\n\
+  %n   File name\n\
+  %o   IO block size\n\
+  %s   Total size, in bytes\n\
+  %T   Minor device type in hex\n\
+  %t   Major device type in hex\n\
 "), stdout);
       fputs (_("\
-  %U - User name of owner\n\
-  %u - User ID of owner\n\
-  %X - Time of last access as seconds since Epoch\n\
-  %x - Time of last access\n\
-  %Y - Time of last modification as seconds since Epoch\n\
-  %y - Time of last modification\n\
-  %Z - Time of last change as seconds since Epoch\n\
-  %z - Time of last change\n\
+  %U   User name of owner\n\
+  %u   User ID of owner\n\
+  %X   Time of last access as seconds since Epoch\n\
+  %x   Time of last access\n\
+  %Y   Time of last modification as seconds since Epoch\n\
+  %y   Time of last modification\n\
+  %Z   Time of last change as seconds since Epoch\n\
+  %z   Time of last change\n\
 \n\
 "), stdout);
 
       fputs (_("\
 Valid format sequences for file systems:\n\
 \n\
-  %a - Free blocks available to non-superuser\n\
-  %b - Total data blocks in file system\n\
-  %c - Total file nodes in file system\n\
-  %d - Free file nodes in file system\n\
-  %f - Free blocks in file system\n\
+  %a   Free blocks available to non-superuser\n\
+  %b   Total data blocks in file system\n\
+  %c   Total file nodes in file system\n\
+  %d   Free file nodes in file system\n\
+  %f   Free blocks in file system\n\
 "), stdout);
       fputs (_("\
-  %i - File System id in hex\n\
-  %l - Maximum length of filenames\n\
-  %n - File name\n\
-  %s - Optimal transfer block size\n\
-  %T - Type in human readable form\n\
-  %t - Type in hex\n\
+  %i   File System id in hex\n\
+  %l   Maximum length of filenames\n\
+  %n   File name\n\
+  %s   Optimal transfer block size\n\
+  %T   Type in human readable form\n\
+  %t   Type in hex\n\
 "), stdout);
       printf (_("\nReport bugs to <%s>.\n"), PACKAGE_BUGREPORT);
     }
@@ -836,5 +805,5 @@ main (int argc, char *argv[])
 	do_statfs (argv[i], terse, format);
     }
 
-  exit (EXIT_SUCCESS);
+  exit (G_fail ? EXIT_FAILURE : EXIT_SUCCESS);
 }
